@@ -1,19 +1,31 @@
-import os, subprocess, copy, json, tqdm, time, threading, queue, requests, abc, logging, backoff
+import abc
+import backoff
+import copy
+import json
+import logging
 import multiprocessing as mp
-import pandas as pd
-import numpy as np
-import pyvo as vo
-from astropy.io import ascii
+import os
+import queue
+import requests
+import subprocess
+import threading
+import time
+import tqdm
+
 import astropy.units as u
-from astropy.table import Table
+import matplotlib.pyplot as plt
+import numpy as np
+import pandas as pd
+import pyvo as vo
 from astropy import constants
 from astropy.cosmology import Planck18
-import matplotlib.pyplot as plt
+from astropy.io import ascii
+from astropy.table import Table
 
-from timewise.general import main_logger, cache_dir, plots_dir, output_dir, logger_format, backoff_hndlr
+from timewise.general import cache_dir, plots_dir, output_dir, logger_format, backoff_hndlr
+from timewise.utils import StableTAPService
 
-
-logger = main_logger.getChild(__name__)
+logger = logging.getLogger(__name__)
 
 
 class WISEDataBase(abc.ABC):
@@ -31,7 +43,7 @@ class WISEDataBase(abc.ABC):
     """
 
     service_url = 'https://irsa.ipac.caltech.edu/TAP'
-    service = vo.dal.TAPService(service_url)
+    service = StableTAPService(service_url)
     active_tap_phases = {"QUEUED", "EXECUTING", "RUN", "COMPLETED", "ERROR", "UNKNOWN"}
     running_tap_phases = ["QUEUED", "EXECUTING", "RUN"]
     done_tap_phases = {"COMPLETED", "ABORTED", "ERROR"}
@@ -265,6 +277,24 @@ class WISEDataBase(abc.ABC):
         # END CHUNK MASK        #
         #######################################################################################
 
+    def _start_data_product(self, parent_sample_indices):
+
+        # get all rows in this chunk and columns, specified in the keymap
+        parent_sample_sel = self.parent_sample.df.loc[
+            parent_sample_indices,
+            list(self.parent_sample.default_keymap.values())
+        ]
+
+        # invert the keymap to rename the columns
+        inverse_keymap = {v: k for k, v in self.parent_sample.default_keymap.items()}
+        parent_sample_sel.rename(columns=inverse_keymap, inplace=True)
+        parent_sample_sel.set_index(parent_sample_sel.index.astype(str), inplace=True)
+
+        # save to data_product
+        data_product = parent_sample_sel.to_dict(orient="index")
+
+        return data_product
+
     @staticmethod
     def get_db_name(table_name, nice=False):
         """
@@ -290,20 +320,27 @@ class WISEDataBase(abc.ABC):
 
     def match_all_chunks(self,
                          table_name="AllWISE Source Catalog",
-                         save_when_done=True):
+                         save_when_done=True,
+                         additional_columns=None):
         """
-        Some descritopn
+        Match the parent sample to a WISE catalogue and add the result to the parent sample.
 
         :param table_name: The name of the table you want to match against
         :type table_name: str
         :param save_when_done: save the parent sample dataframe with the matching info when done
         :type save_when_done: bool
+        :param additional_columns: optional, additional columns to add to the matching table
+        :type additional_columns: list
         :return:
         """
 
         logger.info(f'matching all chunks to {table_name}')
+
+        if additional_columns is None:
+            additional_columns = []
+
         for i in range(self.n_chunks):
-            self._match_single_chunk(i, table_name)
+            self._match_single_chunk(i, table_name, additional_columns)
 
         _dupe_mask = self._get_dubplicated_wise_id_mask()
 
@@ -373,9 +410,15 @@ class WISEDataBase(abc.ABC):
         else:
             return 0
 
-    def _match_to_wise(self, in_filename, out_filename, mask, table_name,
-                       # remove_when_done=True,
-                       N_retries=10, **gator_kwargs):
+    def _match_to_wise(
+            self,
+            in_filename,
+            out_filename,
+            mask,
+            table_name,
+            N_retries=10,
+            **gator_kwargs
+    ):
         selected_parent_sample = copy.copy(
             self.parent_sample.df.loc[mask, [self.parent_ra_key, self.parent_dec_key]])
         selected_parent_sample.rename(columns={self.parent_dec_key: 'dec',
@@ -418,25 +461,35 @@ class WISEDataBase(abc.ABC):
             finally:
                 N_retries -= 1
 
-    def _match_single_chunk(self, chunk_number, table_name):
+    def _match_single_chunk(self, chunk_number, table_name, additional_columns=None):
         """
         Match the parent sample to WISE
 
         :param chunk_number: number of the declination chunk
         :type chunk_number: int
         :param table_name: optional, WISE table to match to, default is AllWISE Source Catalog
-        :type table_name: str
+        :type table_name: str,
+        :param additional_columns: optional, additional columns to be added to the parent sample
+        :type additional_columns: list
         """
 
         dec_intervall_mask = self.chunk_map == chunk_number
         logger.debug(f"Any selected: {np.any(dec_intervall_mask)}")
         _parent_sample_declination_band_file = os.path.join(self.cache_dir, f"parent_sample_chunk{chunk_number}.xml")
         _output_file = os.path.join(self.cache_dir, f"parent_sample_chunk{chunk_number}.tbl")
+
+        additional_keys = (
+            "," + ",".join(additional_columns)
+            if (additional_columns is not None) and (len(additional_columns) > 0)
+            else ""
+        )
+
         gator_res = self._match_to_wise(
             in_filename=_parent_sample_declination_band_file,
             out_filename=_output_file,
             mask=dec_intervall_mask,
-            table_name=table_name
+            table_name=table_name,
+            additional_keys=additional_keys,
         )
 
         for fn in [_parent_sample_declination_band_file, _output_file]:
@@ -458,6 +511,18 @@ class WISEDataBase(abc.ABC):
             self.parent_wise_source_id_key
         ] = list(gator_res["cntr"])
 
+        if len(additional_columns) > 0:
+            for col in additional_columns:
+                logger.debug(f"inserting {col}")
+
+                if col not in self.parent_sample.df.columns:
+                    self.parent_sample.df[col] = np.nan
+
+                self.parent_sample.df.loc[
+                    dec_intervall_mask,
+                    col
+                ] = list(gator_res[col])
+
         _no_match_mask = self.parent_sample.df[self.parent_sample_wise_skysep_key].isna() & dec_intervall_mask
         for k, default in self.parent_sample_default_entries.items():
             self.parent_sample.df.loc[_no_match_mask, k] = default
@@ -482,7 +547,8 @@ class WISEDataBase(abc.ABC):
     ###################################
 
     def get_photometric_data(self, tables=None, perc=1, wait=0, service=None, nthreads=100,
-                             chunks=None, overwrite=True, remove_chunks=False, query_type='positional'):
+                             chunks=None, overwrite=True, remove_chunks=False, query_type='positional',
+                             skip_download=False):
         """
         Load photometric data from the IRSA server for the matched sample. The result will be saved under
 
@@ -506,6 +572,8 @@ class WISEDataBase(abc.ABC):
         :type chunks: list-like
         :param query_type: 'positional': query photometry based on distance from object, 'by_allwise_id': select all photometry points within a radius of 50 arcsec with the corresponding AllWISE ID
         :type query_type: str
+        :param skip_download: if `True` skip downloading and only do binning
+        :type skip_download: bool
         """
 
         mag = True
@@ -535,35 +603,53 @@ class WISEDataBase(abc.ABC):
         if (query_type == 'by_allwise_id') and (service == 'gator'):
             raise ValueError(f"Query type 'by_allwise_id' only implemented for service 'tap'!")
 
-        logger.debug(f"Getting {perc * 100:.2f}% of lightcurve chunks ({len(chunks)}) via {service} "
-                     f"in {'magnitude' if mag else ''} {'flux' if flux else ''} "
-                     f"from {tables}")
+        if not skip_download:
 
-        if service == 'tap':
-            self._query_for_photometry(tables, chunks, wait, mag, flux, nthreads, query_type)
+            logger.debug(f"Getting {perc * 100:.2f}% of lightcurve chunks ({len(chunks)}) via {service} "
+                         f"in {'magnitude' if mag else ''} {'flux' if flux else ''} "
+                         f"from {tables}")
 
-        elif service == 'gator':
-            self._query_for_photometry_gator(tables, chunks, mag, flux, nthreads)
+            if service == 'tap':
+                self._query_for_photometry(tables, chunks, wait, mag, flux, nthreads, query_type)
+
+            elif service == 'gator':
+                self._query_for_photometry_gator(tables, chunks, mag, flux, nthreads)
+
+        else:
+            logger.info("skipping download, assume data is already downloaded.")
 
         self._select_individual_lightcurves_and_bin(service=service, chunks=chunks)
         for c in chunks:
             self.calculate_metadata(service=service, chunk_number=c, overwrite=True)
 
-        self._combine_lcs(service=service, overwrite=overwrite, remove=remove_chunks)
-        self._combine_metadata(service=service, overwrite=overwrite, remove=remove_chunks)
+        self._combine_data_products(service=service, remove=remove_chunks, overwrite=overwrite)
+
+    def _data_product_filename(self, service, chunk_number=None, jobID=None):
+
+        n = "timewise_data_product_"
 
-    def _lightcurve_filename(self, service, chunk_number=None, jobID=None):
         if (chunk_number is None) and (jobID is None):
-            return os.path.join(self.lightcurve_dir, f"binned_lightcurves_{service}.json")
+            return os.path.join(self.lightcurve_dir, f"{n}{service}.json")
         else:
-            fn = f"binned_lightcurves_{service}{self._split_chunk_key}{chunk_number}"
+            fn = f"{n}{service}{self._split_chunk_key}{chunk_number}"
             if (chunk_number is not None) and (jobID is None):
                 return os.path.join(self._cache_photometry_dir, fn + ".json")
             else:
                 return os.path.join(self._cache_photometry_dir, fn + f"_{jobID}.json")
 
-    def _load_lightcurves(self, service, chunk_number=None, jobID=None, return_filename=False):
-        fn = self._lightcurve_filename(service, chunk_number, jobID)
+    def load_data_product(self, service, chunk_number=None, jobID=None, return_filename=False):
+        """
+        Load data product from disk
+
+        :param service: service used to download data ('tap' or 'gator')
+        :type service: str
+        :param chunk_number: chunk number to load, if None load combined file for this service
+        :type chunk_number: int, optional
+        :param jobID: jobID to load, if None load the combined file for this chunk
+        :type jobID: int, optional
+        :param return_filename: return filename of data product, defaults to False
+        """
+        fn = self._data_product_filename(service, chunk_number, jobID)
         logger.debug(f"loading {fn}")
         try:
             with open(fn, "r") as f:
@@ -574,40 +660,31 @@ class WISEDataBase(abc.ABC):
         except FileNotFoundError:
             logger.warning(f"No file {fn}")
 
-    def _save_lightcurves(self, lcs, service, chunk_number=None, jobID=None, overwrite=False):
-        fn = self._lightcurve_filename(service, chunk_number, jobID)
-        logger.debug(f"saving {len(lcs)} new lightcurves to {fn}")
+    def _save_data_product(self, data_product, service, chunk_number=None, jobID=None, overwrite=False):
+        fn = self._data_product_filename(service, chunk_number, jobID)
+        logger.debug(f"saving {len(data_product)} new lightcurves to {fn}")
 
-        if fn == self._lightcurve_filename(service):
-            self._cached_final_products['lightcurves'][service] = lcs
+        if fn == self._data_product_filename(service):
+            self._cached_final_products['lightcurves'][service] = data_product
 
         if not overwrite:
             try:
-                old_lcs = self._load_lightcurves(service=service, chunk_number=chunk_number, jobID=jobID)
-                logger.debug(f"Found {len(old_lcs)}. Combining")
-                lcs = lcs.update(old_lcs)
+                old_data_product = self.load_data_product(service=service, chunk_number=chunk_number, jobID=jobID)
+                logger.debug(f"Found {len(old_data_product)}. Combining")
+                data_product = data_product.update(old_data_product)
             except FileNotFoundError as e:
                 logger.info(f"FileNotFoundError: {e}. Making new binned lightcurves.")
 
         with open(fn, "w") as f:
-            json.dump(lcs, f)
+            json.dump(data_product, f, indent=4)
 
-    def load_binned_lcs(self, service):
-        """Loads the binned lightcurves. For any int `ID` the lightcurves can convieniently read into a pandas.DataFrame
-        via
-
-            lc = pandas.DataFrame.from_dict(json_dictionary[ID])
-
-        :param service: the service with which the lightcuvres were downloaded
-        :type service: str
-        :return: the binned lightcurves
-        :rtype: dict
-        """
-        if service not in self._cached_final_products['lightcurves']:
-            self._cached_final_products['lightcurves'][service] = self._load_lightcurves(service)
-        return self._cached_final_products['lightcurves'][service]
-
-    def _combine_lcs(self, service=None, chunk_number=None, remove=False, overwrite=False):
+    def _combine_data_products(
+            self,
+            service=None,
+            chunk_number=None,
+            remove=False,
+            overwrite=False
+    ):
         if not service:
             logger.info("Combining all lightcuves collected with all services")
             itr = ['service', ['gator', 'tap']]
@@ -617,7 +694,7 @@ class WISEDataBase(abc.ABC):
             itr = ['chunk_number', range(self.n_chunks)]
             kwargs = {'service': service}
         elif chunk_number is not None:
-            logger.info(f"Combining all lightcurves collected eith {service} for chunk {chunk_number}")
+            logger.info(f"Combining all lightcurves collected with {service} for chunk {chunk_number}")
             itr = ['jobID',
                    list(self.clusterJob_chunk_map.index[self.clusterJob_chunk_map.chunk_number == chunk_number])]
             kwargs = {'service': service, 'chunk_number': chunk_number}
@@ -626,11 +703,13 @@ class WISEDataBase(abc.ABC):
 
         lcs = None
         fns = list()
+        missing_files = 0
         for i in itr[1]:
             kw = dict(kwargs)
             kw[itr[0]] = i
             kw['return_filename'] = True
-            res = self._load_lightcurves(**kw)
+            res = self.load_data_product(**kw)
+
             if not isinstance(res, type(None)):
                 ilcs, ifn = res
                 fns.append(ifn)
@@ -639,11 +718,28 @@ class WISEDataBase(abc.ABC):
                 else:
                     lcs.update(ilcs)
 
-        self._save_lightcurves(lcs, service=service, chunk_number=chunk_number, overwrite=overwrite)
+            else:
+                missing_files += 1
 
-        if remove:
-            for fn in tqdm.tqdm(fns, desc="removing files"):
-                os.remove(fn)
+            if missing_files > 0:
+                msg = f"Missing {missing_files} for {service}"
+                if chunk_number is not None:
+                    msg += f" chunk {chunk_number}"
+                msg += "! Not saving data product"
+                logger.warning(msg)
+                break
+
+        if missing_files == 0:
+            self._save_data_product(lcs, service=service, chunk_number=chunk_number, overwrite=overwrite)
+
+            if remove:
+                for fn in tqdm.tqdm(fns, desc="removing files"):
+                    os.remove(fn)
+
+            return True
+
+        else:
+            return False
 
     # ----------------------------------------------------------------------------------- #
     # START using GATOR to get photometry        #
@@ -844,7 +940,7 @@ class WISEDataBase(abc.ABC):
                                                         f"of {t} "
                                                         f"after {N_tries} attempts")
             try:
-                job = WISEDataBase.service.submit_job(qstring, uploads={'ids': Table(tab_d)})
+                job = self.service.submit_job(qstring, uploads={'ids': Table(tab_d)})
                 job.run()
 
                 if isinstance(job.phase, type(None)):
@@ -980,8 +1076,7 @@ class WISEDataBase(abc.ABC):
             self.tap_jobs[t] = dict()
             for i in chunks:
                 self._submit_job_to_TAP(i, t, mag, flux, query_type)
-                logger.info("waiting for 2 minutes")
-                time.sleep(2*60)
+                time.sleep(5)
 
         logger.info(f'added {self.queue.qsize()} tasks to queue')
         logger.info(f"wait for {wait} hours to give tap_jobs some time")
@@ -1029,7 +1124,15 @@ class WISEDataBase(abc.ABC):
         else:
             r = list(map(self._subprocess_select_and_bin, service_list, chunk_list))
 
-    def _get_unbinned_lightcurves(self, chunk_number, clear=False):
+    def get_unbinned_lightcurves(self, chunk_number, clear=False):
+        """
+        Get the unbinned lightcurves for a given chunk number.
+
+        :param chunk_number: int
+        :type chunk_number: int
+        :param clear: remove files after loading, defaults to False
+        :type clear: bool, optional
+        """
         # load only the files for this chunk
         fns = [os.path.join(self._cache_photometry_dir, fn)
                for fn in os.listdir(self._cache_photometry_dir)
@@ -1037,6 +1140,10 @@ class WISEDataBase(abc.ABC):
                 f"{self._split_chunk_key}{chunk_number}.csv"
             ))]
         logger.debug(f"chunk {chunk_number}: loading {len(fns)} files for chunk {chunk_number}")
+
+        if len(fns) == 0:
+            raise ValueError(f"No unbinned lightcurves found for chunk {chunk_number}!")
+
         lightcurves = pd.concat([pd.read_csv(fn) for fn in fns])
 
         if clear:
@@ -1048,10 +1155,7 @@ class WISEDataBase(abc.ABC):
     def _subprocess_select_and_bin(self, service, chunk_number=None, jobID=None):
         # run through the ids and bin the lightcurves
         if service == 'tap':
-            lightcurves = self._get_unbinned_lightcurves(
-                chunk_number,
-                clear=self.clear_unbinned_photometry_when_binning
-            )
+            lightcurves = self.get_unbinned_lightcurves(chunk_number, clear=self.clear_unbinned_photometry_when_binning)
         elif service == 'gator':
             lightcurves = self._get_unbinned_lightcurves_gator(
                 chunk_number,
@@ -1067,8 +1171,13 @@ class WISEDataBase(abc.ABC):
 
         logger.debug(f"chunk {chunk_number}: going through {len(indices)} IDs")
 
-        binned_lcs = dict()
-        for parent_sample_entry_id in tqdm.tqdm(indices):
+        data_product = self.load_data_product(service=service, chunk_number=chunk_number, jobID=jobID)
+
+        if data_product is None:
+            logger.info(f"Starting data product for {len(indices)} indices.")
+            data_product = self._start_data_product(parent_sample_indices=indices)
+
+        for parent_sample_entry_id in tqdm.tqdm(indices, desc="binning"):
             m = lightcurves[self._tap_orig_id_key] == parent_sample_entry_id
             lightcurve = lightcurves[m]
 
@@ -1077,10 +1186,10 @@ class WISEDataBase(abc.ABC):
                 continue
 
             binned_lc = self.bin_lightcurve(lightcurve)
-            binned_lcs[f"{int(parent_sample_entry_id)}"] = binned_lc.to_dict()
+            data_product[str(int(parent_sample_entry_id))]["timewise_lightcurve"] = binned_lc.to_dict()
 
-        logger.debug(f"chunk {chunk_number}: saving {len(binned_lcs.keys())} binned lcs")
-        self._save_lightcurves(binned_lcs, service=service, chunk_number=chunk_number, jobID=jobID, overwrite=True)
+        logger.debug(f"chunk {chunk_number}: saving {len(data_product.keys())} binned lcs")
+        self._save_data_product(data_product, service=service, chunk_number=chunk_number, jobID=jobID, overwrite=True)
 
     # ---------------------------------------- #
     # END using TAP to get photometry          #
@@ -1227,10 +1336,10 @@ class WISEDataBase(abc.ABC):
         :param service: The service with which the lightcurves were downloaded
         :type service: str
         """
-        lcs = self.load_binned_lcs(service=service)
-        for i, lc in tqdm.tqdm(lcs.items(), desc='adding flux densities'):
-            lcs[i] = self.add_flux_density(
-                lc,
+        data_product = self.load_data_product(service=service)
+        for i, i_data_product in tqdm.tqdm(data_product.items(), desc='adding flux densities'):
+            data_product[i]["timewise_lightcurve"] = self.add_flux_density(
+                i_data_product["timewise_lightcurve"],
                 mag_key=f'{self.mean_key}{self.mag_key_ext}',
                 emag_key=f'{self.mag_key_ext}{self.rms_key}',
                 mag_ul_key=f'{self.mag_key_ext}{self.upper_limit_key}',
@@ -1238,7 +1347,7 @@ class WISEDataBase(abc.ABC):
                 ef_key=f'{self.flux_density_key_ext}{self.rms_key}',
                 f_ul_key=f'{self.flux_density_key_ext}{self.upper_limit_key}'
             ).to_dict()
-        self._save_lightcurves(lcs, service=service, overwrite=True)
+        self._save_data_product(data_product, service=service, overwrite=True)
 
     # ---------------------------------------------------- #
     # END converting to flux densities                     #
@@ -1292,7 +1401,7 @@ class WISEDataBase(abc.ABC):
         return lightcurve
 
     def add_luminosity_to_saved_lightcurves(self, service, redshift_key=None, distance_key=None):
-        """Add luminosities to all lightcurves, calculated from flux desnities and distance or redshift
+        """Add luminosities to all lightcurves, calculated from flux densities and distance or redshift
 
         :param service: the service with which the lightcurves were downloaded
         :type service: str
@@ -1305,8 +1414,8 @@ class WISEDataBase(abc.ABC):
         if (not redshift_key) and (not distance_key):
             raise ValueError('Either distance key or redshift key has to be given!')
 
-        lcs = self.load_binned_lcs(service=service)
-        for i, lc in tqdm.tqdm(lcs.items(), desc='adding luminosities'):
+        data_product = self.load_data_product(service=service)
+        for i, i_data_product in tqdm.tqdm(data_product.items(), desc='adding luminosities'):
             parent_sample_idx = int(i.split('_')[0])
             info = self.parent_sample.df.loc[parent_sample_idx]
 
@@ -1317,8 +1426,8 @@ class WISEDataBase(abc.ABC):
                 distance = None
                 redshift = info[redshift_key]
 
-            lcs[i] = self._add_luminosity(
-                pd.DataFrame.from_dict(lc),
+            data_product[i]["timewise_lightcurve"] = self._add_luminosity(
+                pd.DataFrame.from_dict(i_data_product["timewise_lightcurve"]),
                 f_key     = self.mean_key + self.flux_density_key_ext,
                 ef_key    = self.flux_density_key_ext + self.rms_key,
                 f_ul_key  = self.flux_density_key_ext + self.upper_limit_key,
@@ -1328,7 +1437,7 @@ class WISEDataBase(abc.ABC):
                 redshift  = redshift,
                 distance  = distance
             ).to_dict()
-        self._save_lightcurves(lcs, service=service, overwrite=True)
+        self._save_data_product(data_product, service=service, overwrite=True)
 
     # ---------------------------------------------------- #
     # END converting to luminosity                         #
@@ -1367,21 +1476,21 @@ class WISEDataBase(abc.ABC):
         """
 
         logger.debug(f"loading binned lightcurves")
-        lcs = self.load_binned_lcs(service)
-        _get_unbinned_lcs_fct = self._get_unbinned_lightcurves if service == 'tap' else self._get_unbinned_lightcurves_gator
+        data_product = self.load_data_product(service)
+        _get_unbinned_lcs_fct = self.get_unbinned_lightcurves if service == 'tap' else self._get_unbinned_lightcurves_gator
 
         wise_id = self.parent_sample.df.loc[int(parent_sample_idx), self.parent_wise_source_id_key]
         if isinstance(wise_id, float) and not np.isnan(wise_id):
             wise_id = int(wise_id)
         logger.debug(f"{wise_id} for {parent_sample_idx}")
 
-        lc = pd.DataFrame.from_dict(lcs[f"{int(parent_sample_idx)}"])
+        lc = pd.DataFrame.from_dict(data_product[str(int(parent_sample_idx))]["timewise_lightcurve"])
 
         if plot_unbinned:
             _chunk_number = self._get_chunk_number(parent_sample_index=parent_sample_idx)
 
             if service == 'tap':
-                unbinned_lcs = self._get_unbinned_lightcurves(_chunk_number)
+                unbinned_lcs = self.get_unbinned_lightcurves(_chunk_number)
 
             else:
                 unbinned_lcs = self._get_unbinned_lightcurves_gator(_chunk_number)
@@ -1423,7 +1532,7 @@ class WISEDataBase(abc.ABC):
                                 label=f"{b} unbinned", ls='', marker='o', c=self.band_plot_colors[b], markersize=4,
                                 alpha=0.3)
             except KeyError as e:
-                logger.warning(e)
+                raise KeyError(f"Could not find brightness key {e}!")
 
         if lum_key == 'mag':
             ylim = ax.get_ylim()
@@ -1450,58 +1559,6 @@ class WISEDataBase(abc.ABC):
     #  START CALCULATE METADATA         #
     #####################################
 
-    def _metadata_filename(self, service, chunk_number=None, jobID=None):
-        if (chunk_number is None) and (jobID is None):
-            return os.path.join(self.lightcurve_dir, f'metadata_{service}.json')
-        elif (chunk_number is not None) and (jobID is None):
-            return os.path.join(self.cache_dir, f'metadata_{service}{self._split_chunk_key}{chunk_number}.json')
-        elif (chunk_number is not None) and (jobID is not None):
-            return os.path.join(self.cache_dir, f'metadata_{service}{self._split_chunk_key}{chunk_number}_job{jobID}.json')
-        else:
-            raise NotImplementedError
-
-    def _load_metadata(self, service, chunk_number=None, jobID=None, return_filename=False):
-        fn = self._metadata_filename(service, chunk_number, jobID)
-        try:
-            logger.debug(f"loading {fn}")
-            with open(fn, "r") as f:
-                metadata = json.load(f)
-            if return_filename:
-                return metadata, fn
-            return metadata
-        except FileNotFoundError:
-            logger.warning(f"No file {fn}")
-
-    def _save_metadata(self, metadata, service, chunk_number=None, jobID=None, overwrite=False):
-        fn = self._metadata_filename(service, chunk_number, jobID)
-
-        if fn == self._metadata_filename(service):
-            self._cached_final_products['metadata'][service] = metadata
-
-        if not overwrite:
-            try:
-                old_metadata = self._load_metadata(service=service, chunk_number=chunk_number, jobID=jobID)
-                logger.debug(f"Found {len(old_metadata)}. Combining")
-                metadata = metadata.update(old_metadata)
-            except FileNotFoundError as e:
-                logger.info(f"FileNotFoundError: {e}. Making new metadata.")
-
-        logger.debug(f'saving under {fn}')
-        with open(fn, "w") as f:
-            json.dump(metadata, f)
-
-    def load_metadata(self, service):
-        """Load the metadata
-
-        :param service: The service with which the lightcurves were downloaded
-        :type service: str
-        :return: the metadata
-        :rtype: dict
-        """
-        if not service in self._cached_final_products['metadata']:
-            self._cached_final_products['metadata'][service] = self._load_metadata(service)
-        return self._cached_final_products['metadata'][service]
-
     def calculate_metadata(self, service, chunk_number=None, jobID=None, overwrite=True):
         """Calculates the metadata for all downloaded lightcurves.
          Results will be saved under
@@ -1517,47 +1574,14 @@ class WISEDataBase(abc.ABC):
         :param overwrite: overwrite existing metadata file
         :type overwrite: bool
         """
-        lcs = self._load_lightcurves(service, chunk_number, jobID)
-        metadata = self.calculate_metadata_single(lcs)
-        self._save_metadata(metadata, service, chunk_number, jobID, overwrite=overwrite)
+        data_product = self.load_data_product(service, chunk_number, jobID)
+        for ID, i_data_product in tqdm.tqdm(data_product.items(), desc="calculating metadata"):
+            if "timewise_lightcurve" in i_data_product:
+                lc = pd.DataFrame.from_dict(i_data_product["timewise_lightcurve"])
+                metadata = self.calculate_metadata_single(lc)
+                data_product[ID]["timewise_metadata"] = metadata
 
-    def _combine_metadata(self, service=None, chunk_number=None, remove=False, overwrite=False):
-        if not service:
-            logger.info("Combining all metadata collected with all services")
-            itr = ['service', ['gator', 'tap']]
-            kwargs = {}
-        elif chunk_number is None:
-            logger.info(f"Combining all metadata collected with {service}")
-            itr = ['chunk_number', range(self.n_chunks)]
-            kwargs = {'service': service}
-        elif chunk_number is not None:
-            logger.info(f"Combining all metadata collected with {service} for chunk {chunk_number}")
-            itr = ['jobID',
-                   list(self.clusterJob_chunk_map.index[self.clusterJob_chunk_map.chunk_number == chunk_number])]
-            kwargs = {'service': service, 'chunk_number': chunk_number}
-        else:
-            raise NotImplementedError
-
-        metadata = None
-        fns = list()
-        for i in itr[1]:
-            kw = dict(kwargs)
-            kw[itr[0]] = i
-            kw['return_filename'] = True
-            res = self._load_metadata(**kw)
-            if not isinstance(res, type(None)):
-                imetadata, ifn = res
-                fns.append(ifn)
-                if isinstance(metadata, type(None)):
-                    metadata = dict(imetadata)
-                else:
-                    metadata.update(imetadata)
-
-        self._save_metadata(metadata, service=service, chunk_number=chunk_number, overwrite=overwrite)
-
-        if remove:
-            for fn in tqdm.tqdm(fns, desc='removing files'):
-                os.remove(fn)
+        self._save_data_product(data_product, service, chunk_number, jobID, overwrite=overwrite)
 
     @abc.abstractmethod
     def calculate_metadata_single(self, lcs):
